import importlib
from collections.abc import Callable, Iterator, Sequence
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Annotated

import dagster as dg
from dagster import AssetKey, AssetSpec, Component, ComponentLoadContext, Resolvable, Resolver
from dagster._annotations import public
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster.components.resolved.context import ResolutionContext
from dagster.components.scaffold.scaffold import scaffold_with
from dagster.components.utils.translation import (
    ComponentTranslator,
    TranslationFn,
    TranslationFnResolver,
    create_component_translator_cls,
)
from dlt import Pipeline
from dlt.extract.source import DltSource

from dagster_dlt.asset_decorator import dlt_assets
from dagster_dlt.components.dlt_load_collection.scaffolder import DltLoadCollectionScaffolder
from dagster_dlt.translator import DagsterDltTranslator, DltResourceTranslatorData

if TYPE_CHECKING:
    from dagster_dlt import DagsterDltResource


def _load_object_from_python_path(resolution_context: ResolutionContext, path: str):
    """Loads a Python object from the given import path, accepting
    relative paths.

    For example, '.foo_module.bar_object' will find the relative module
    'foo_module' and return 'bar_object'.
    """
    context = ComponentLoadContext.from_resolution_context(resolution_context)

    if path.startswith("."):
        path = f"{context.defs_relative_module_name(context.path)}{path}"
    module_name, object_name = path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, object_name)


@dataclass
class DltLoadSpecModel(Resolvable):
    """Represents a single dlt load, a combination of pipeline and source."""

    pipeline: Annotated[
        Pipeline,
        Resolver(_load_object_from_python_path, model_field_type=str),
    ]
    source: Annotated[
        DltSource,
        Resolver(
            _load_object_from_python_path,
            model_field_type=str,
        ),
    ]
    translation: (
        Annotated[
            TranslationFn[DltResourceTranslatorData],
            TranslationFnResolver[DltResourceTranslatorData](
                lambda data: {"resource": data.resource, "pipeline": data.pipeline}
            ),
        ]
        | None
    ) = None


@public
@scaffold_with(DltLoadCollectionScaffolder)
@dataclass
class DltLoadCollectionComponent(Component, Resolvable):
    """Expose one or more dlt loads to Dagster as assets."""

    loads: Sequence[DltLoadSpecModel]

    @property
    def dlt_pipeline_resource(self) -> "DagsterDltResource":
        from dagster_dlt import DagsterDltResource

        return DagsterDltResource()

    @cached_property
    def _base_translator(self) -> DagsterDltTranslator:
        return DagsterDltTranslator()

    @public
    def get_asset_spec(self, data: DltResourceTranslatorData) -> AssetSpec:
        """Generates an AssetSpec for a given dlt resource.

        This method can be overridden in a subclass to customize how dlt resources are
        converted to Dagster asset specs. By default, it delegates to the configured
        DagsterDltTranslator.

        Args:
            data: The DltResourceTranslatorData containing information about the dlt source
                and resource being loaded

        Returns:
            An AssetSpec that represents the dlt resource as a Dagster asset

        Example:
            Override this method to add custom tags based on resource properties:

            .. code-block:: python

                from dagster_dlt import DltLoadCollectionComponent
                from dagster import AssetSpec

                class CustomDltLoadCollectionComponent(DltLoadCollectionComponent):
                    def get_asset_spec(self, data):
                        base_spec = super().get_asset_spec(data)
                        return base_spec.replace_attributes(
                            tags={
                                **base_spec.tags,
                                "source": data.source_name,
                                "resource": data.resource_name
                            }
                        )
        """
        return self._base_translator.get_asset_spec(data)

    def build_defs(self, context: ComponentLoadContext) -> dg.Definitions:
        output = []
        for load in self.loads:
            translator = DltComponentTranslator(self, load)

            @dlt_assets(
                dlt_source=load.source,
                dlt_pipeline=load.pipeline,
                name=f"dlt_assets_{load.source.name}_{load.pipeline.dataset_name}",
                dagster_dlt_translator=translator,
            )
            def dlt_assets_def(context: AssetExecutionContext):
                yield from self.execute(context, self.dlt_pipeline_resource)

            output.append(dlt_assets_def)

        return dg.Definitions(assets=output)

    @public
    def execute(
        self, context: AssetExecutionContext, dlt_pipeline_resource: "DagsterDltResource"
    ) -> Iterator:
        """Executes the dlt pipeline for the selected resources.

        This method can be overridden in a subclass to customize the pipeline execution behavior,
        such as adding custom logging, validation, or error handling.

        Args:
            context: The asset execution context provided by Dagster
            dlt_pipeline_resource: The DagsterDltResource used to run the dlt pipeline

        Yields:
            Events from the dlt pipeline execution (e.g., AssetMaterialization, MaterializeResult)

        Example:
            Override this method to add custom logging during pipeline execution:

            .. code-block:: python

                from dagster_dlt import DltLoadCollectionComponent
                from dagster import AssetExecutionContext

                class CustomDltLoadCollectionComponent(DltLoadCollectionComponent):
                    def execute(self, context, dlt_pipeline_resource):
                        context.log.info("Starting dlt pipeline execution")
                        yield from super().execute(context, dlt_pipeline_resource)
                        context.log.info("dlt pipeline execution completed")
        """
        yield from dlt_pipeline_resource.run(context=context)


class DltComponentTranslator(
    create_component_translator_cls(DltLoadCollectionComponent, DagsterDltTranslator),
    ComponentTranslator[DltLoadCollectionComponent],
):
    def __init__(self, component: "DltLoadCollectionComponent", load_spec: "DltLoadSpecModel"):
        self._component = component
        self._load_spec = load_spec

    def get_asset_spec(self, data: DltResourceTranslatorData) -> AssetSpec:
        table_name = data.resource.table_name
        if isinstance(table_name, Callable):
            table_name = data.resource.name
        prefix = (
            [data.pipeline.dataset_name] if data.pipeline and data.pipeline.dataset_name else []
        )
        base_asset_spec = (
            super().get_asset_spec(data).replace_attributes(key=AssetKey(prefix + [table_name]))
        )
        if self._load_spec.translation is None:
            return base_asset_spec
        else:
            return self._load_spec.translation(base_asset_spec, data)



import ast
import shutil
import subprocess
import textwrap
from collections.abc import Mapping
from pathlib import Path
from typing import NamedTuple

from dagster import Scaffolder, scaffold_component
from dagster._utils import pushd
from dagster.components.scaffold.scaffold import ScaffoldRequest
from dagster.components.utils import check
from pydantic import BaseModel


def _format_file_if_ruff_installed(file_path: Path) -> None:
    if shutil.which("ruff"):
        subprocess.run(["ruff", "format", file_path], check=False)


class PipelineAndSource(NamedTuple):
    pipeline_src: str
    source_src: str


class ParsedPipelineAndSource(NamedTuple):
    imports: list[str]
    pipelines_and_sources: Mapping[str, PipelineAndSource]


def _extract_pipeline_and_source_from_init_file(
    node: ast.FunctionDef,
) -> PipelineAndSource:
    """Given a function body AST node, extracts the source code for the
    local Pipeline and DltSource objects.
    """
    pipeline_src = None
    source_src = None

    for stmt in node.body:
        if isinstance(stmt, ast.Assign):
            target = stmt.targets[0]
            if isinstance(target, ast.Name):
                var_name = target.id
                if var_name == "pipeline":
                    pipeline_src = ast.unparse(stmt)
                elif var_name == "data":
                    source_src = ast.unparse(stmt)

    return PipelineAndSource(check.not_none(pipeline_src), check.not_none(source_src))


def _extract_pipelines_and_sources_from_pipeline_file(
    file_path: Path,
) -> ParsedPipelineAndSource:
    """Process a Python file and generate a new file with pipeline and data definitions."""
    imports = []
    pipelines_and_sources = {}
    source = file_path.read_text()
    tree = ast.parse(source)

    # Create new file content
    new_content = []
    new_content.append('"""Generated pipeline and data definitions."""\n')

    # Add imports from original file
    for node in tree.body:
        if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
            imports.append(ast.unparse(node).replace("from ", "from ."))

    # Process each function
    for node in tree.body:
        if isinstance(node, ast.FunctionDef):
            pipelines_and_sources[node.name.removeprefix("load_")] = (
                _extract_pipeline_and_source_from_init_file(node)
            )

    return ParsedPipelineAndSource(imports, pipelines_and_sources)


def _process_pipeline(src: str) -> str:
    return src.replace(", dev_mode=True", "")


def _construct_pipeline_source_file(
    file_path: Path,
    parsed_pipeline_and_source: ParsedPipelineAndSource,
) -> None:
    """Construct a new pipeline source file from a set of pipeline and source definitions."""
    new_content = []
    for import_line in parsed_pipeline_and_source.imports:
        new_content.append(import_line)

    new_content.append("\n")

    for load_name, (
        pipeline_src,
        source_src,
    ) in parsed_pipeline_and_source.pipelines_and_sources.items():
        new_content.append(source_src.replace("data =", f"{load_name}_source ="))
        new_content.append(
            _process_pipeline(pipeline_src).replace("pipeline =", f"{load_name}_pipeline =")
        )
        new_content.append("")

    file_path.write_text("\n".join(new_content))


class DltScaffolderParams(BaseModel):
    source: str | None = None
    destination: str | None = None


DLT_INIT_FILES_TO_CLEAN_UP = [".gitignore", "requirements.txt", ".dlt"]


class DltLoadCollectionScaffolder(Scaffolder[DltScaffolderParams]):
    @classmethod
    def get_scaffold_params(cls) -> type[DltScaffolderParams]:
        return DltScaffolderParams

    def scaffold(self, request: ScaffoldRequest[DltScaffolderParams]) -> None:
        with pushd(str(request.target_path)):
            Path.cwd().mkdir(parents=True, exist_ok=True)
            # Given source and destination, we can use dlt init to scaffold the source
            # code and some sample pipelines and sources.
            if request.params and request.params.source and request.params.destination:
                # Use subprocess with text input instead of Unix 'yes' command for cross-platform compatibility
                subprocess.run(
                    ["dlt", "init", request.params.source, request.params.destination],
                    input="y\n" * 100,  # Provide enough 'y' responses for any prompts
                    text=True,
                    check=True,
                )
                # dlt init scaffolds a Python file with some example pipelines, nested in functions
                # we extract them into top-level objects which we stash in loads.py as a sample
                examples_python_file = next(Path(".").glob("*.py"))
                examples_python_file.unlink()
                for file in DLT_INIT_FILES_TO_CLEAN_UP:
                    if Path(file).is_dir():
                        shutil.rmtree(file)
                    else:
                        Path(file).unlink()
            elif request.params and (request.params.source or request.params.destination):
                raise ValueError("Must provide neither or both of source and destination")

            Path("loads.py").write_text(
                textwrap.dedent(
                    """
                    import dlt

                    @dlt.source
                    def my_source():
                        @dlt.resource
                        def hello_world():
                            yield "hello, world!"

                        return hello_world

                    my_load_source = my_source()
                    my_load_pipeline = dlt.pipeline({destination})
                    """
                ).format(
                    destination=f'destination="{request.params.destination}"'
                    if request.params and request.params.destination
                    else "",
                )
            )

            _format_file_if_ruff_installed(Path("loads.py"))

        scaffold_component(
            request=request,
            yaml_attributes={
                "loads": [
                    {
                        "source": ".loads.my_load_source",
                        "pipeline": ".loads.my_load_pipeline",
                    }
                ]
            },
        )
